
package simsql.code_generator;

import java.util.*;

/** A relation, with its name, location, and set of attributes. */
class PrologQueryRelation {

    enum Type {
	RELATION, TEMPORARYTABLE
    };

    private String name;
    private String file;
    private Type type;
    private ArrayList<PrologAttributeExp> attributes = new ArrayList<PrologAttributeExp>();
    private ArrayList<ArrayList<PrologLiteralExp>> rows = new ArrayList<ArrayList<PrologLiteralExp>>();

    public PrologQueryRelation(TupleTerm tup, HashMap<AtomTerm, TupleTerm> allAttributes) {

	// check the arity
	tup.checkArity(3);

	// check the type
	if (tup.getAtom().equals(new AtomTerm("relation"))) {
	    
	    // get the name
	    name = ((AtomTerm)tup.getTerm(0)).getValue();
	    
	    // get the file
	    file = ((AtomTerm)tup.getTerm(1)).getValue();
	    
	    // get the set of attributes.
	    for (PrologTerm t: ((ListTerm)tup.getTerm(2))) {
		attributes.add(new PrologAttributeExp(allAttributes.get(t)));
	    }
	    
	    // set the type
	    type = Type.RELATION;
	} else if (tup.getAtom().equals(new AtomTerm("temporaryTable"))) {

	    // get the name and file
	    name = ((AtomTerm)tup.getTerm(0)).getValue();
	    file = name;

	    // get the set of attributes.
	    for (PrologTerm t: ((ListTerm)tup.getTerm(1))) {
		attributes.add(new PrologAttributeExp(allAttributes.get(t)));
	    }

	    // get the set of literals.
	    for (PrologTerm t: ((ListTerm)tup.getTerm(2))) {

		ArrayList<PrologLiteralExp> tx = new ArrayList<PrologLiteralExp>();
		for (PrologTerm tt: (ListTerm)t) {
		    tx.add(new PrologLiteralExp((AtomTerm)tt));
		}

		rows.add(tx);
	    }

	    // set the type.
	    type = Type.TEMPORARYTABLE;
	}
    }

    public Type getType() {
	return type;
    }

    public String getName() {
	return name;
    }

    public String getFile() {
	return file;
    }

    public ArrayList<PrologAttributeExp> getAttributes() {
	return attributes;
    }

    public String toString() {
	return name;
    }

    public ArrayList<ArrayList<PrologLiteralExp>> getRows() {
	return rows;
    }
}
