package simsql.compiler; // package mcdb.catalog;

/**
 * Encapsulates a single attribute, as a (name,type) pair. Also contains the
 * definition of all types.
 * 
 * @author Luis, Bamboo
 */
public class Attribute {

	/* Attribute type definitions/constants */
	public final static int INTEGER = 1;
	public final static int DOUBLE = 2;
	public final static int STRING = 3;
	public final static int DATE = 4;
	public final static int STOCHINT = 5;
	public final static int STOCHDBL = 6;
	public final static int STOCHSTR = 7;
	public final static int STOCHDAT = 8;
	public final static int SEED = 9;
	public final static int BOTTOM = 10;
	public final static int MULTITYPE = 11;
	public final static int TOTALTYPE = 12;
	public final static int UNKNOWN = 13;
	public final static int HYBRID = 1000;

	/** Relation name. Optional. */
	private String relName;

	/** Attribute name */
	private String name;

	/** Attribute type */
	private int type;

	/*------------------------modified by Bamboo-----------------------------*/
	private String domainType;
	private String domainSize;
	/*------------------------ end for VGAttribute------------------------------------------*/

	/*------------------------modified by Bamboo-----------------------------*/
	private int attributeSize;
	private long uniqueValue;
	/*------------------------ end for VGAttribute------------------------------------------*/

	// records whether this is a random attribute
	private boolean isRandom = false;

	/** Public constructor for "regular" attributes */

	public Attribute(String _name, int _type, String _relName, boolean isRandom) {
		name = _name;
		type = _type;
		relName = _relName;
		domainType = null;
		domainSize = null;
		attributeSize = this.getTypeSize(type);
		this.isRandom = isRandom;
	}

	public Attribute(String _name, int _type, String _relName) {
		name = _name;
		type = _type;
		relName = _relName;
		domainType = null;
		domainSize = null;
		attributeSize = this.getTypeSize(type);
	}

	/* Support VGAttribute */
	public Attribute(String _name, int _type, String _relName,
			String domainType, String domainSize, int attributeSize,
			boolean isRandom) {
		name = _name;
		type = _type;
		relName = _relName;
		this.domainType = domainType;
		this.domainSize = domainSize;
		this.attributeSize = attributeSize;
		this.isRandom = isRandom;
	}

	public Attribute(String _name, int _type, String _relName,
			int _attributeSize, long _uniqueValue) {
		name = _name;
		type = _type;
		relName = _relName;
		attributeSize = _attributeSize;
		uniqueValue = _uniqueValue;
	}

	/** Public constructor, without relation name */
	public Attribute(String _name, int _type) {
		name = _name;
		type = _type;
		relName = "";
	}

	public Attribute(String nameIn, int typeIn, int attSizeIn, String tableIn) {
		name = nameIn;
		type = typeIn;
		attributeSize = attSizeIn;
		relName = tableIn;
	}

	/** Returns the attribute name */
	public String getName() {
		return (name);
	}

	/** Tells us if this guy is random */
	public boolean getIsRandom() {
		return isRandom;
	}

	/** Returns the attribute type */
	public int getType() {
		return (type);
	}

	/** Returns the relation name */
	public String getRelName() {
		return (relName);
	}

	/** Sets the attribute name */
	public void setName(String _name) {
		name = _name;
	}

	/** Sets the attribute type */
	public void setType(int _type) {
		type = _type;
	}

	/** Sets the relation name */
	public void setRelName(String _relName) {
		relName = _relName;
	}

	/**
	 * @return the domainType
	 */
	public String getDomainType() {
		return domainType;
	}

	/**
	 * @param domainType
	 *            the domainType to set
	 */
	public void setDomainType(String domainType) {
		this.domainType = domainType;
	}

	/**
	 * @return the domainSize
	 */
	public String getDomainSize() {
		return domainSize;
	}

	/**
	 * @param domainSize
	 *            the domainSize to set
	 */
	public void setDomainSize(String domainSize) {
		this.domainSize = domainSize;
	}

	/** Returns a string representation, for printf() */
	public String toString() {
		return (relName + "." + name + ":" + type + ":" + uniqueValue);
	}

	/** Returns the entire name of the attribute */
	public String getFullName() {
		return (relName + "." + name);
	}

	/**
	 * @return the attributeSize
	 */
	public int getAttributeSize() {
		return attributeSize;
	}

	/**
	 * @param attributeSize
	 *            the attributeSize to set
	 */
	public void setAttributeSize(int attributeSize) {
		this.attributeSize = attributeSize;
	}

	/**
	 * @return the uniqueValue
	 */
	public long getUniqueValue() {
		return uniqueValue;
	}

	/**
	 * @param uniqueValue
	 *            the uniqueValue to set
	 */
	public void setUniqueValue(long uniqueValue) {
		this.uniqueValue = uniqueValue;
	}

	public String getTypeStr() {
		switch (type) {
		case INTEGER:
			return ("integer");
		case DOUBLE:
			return ("double");
		case STRING:
			return ("string");
		case DATE:
			return ("date");
		case STOCHINT:
			return ("stochint");
		case STOCHDBL:
			return ("stochdbl");
		case STOCHSTR:
			return ("stochstr");
		case STOCHDAT:
			return ("stochdat");
		case SEED:
			return ("seed");
		case BOTTOM:
			return ("bottom");
		default:
			return ("unknown");
		}
	}

	public static int getStringType(String str) {

		if (str.equalsIgnoreCase("integer"))
			return INTEGER;

		if (str.equalsIgnoreCase("double"))
			return DOUBLE;

		if (str.equalsIgnoreCase("string"))
			return STRING;

		if (str.equalsIgnoreCase("date"))
			return DATE;

		if (str.equalsIgnoreCase("stochint"))
			return STOCHINT;

		if (str.equalsIgnoreCase("stochdbl"))
			return STOCHDBL;

		if (str.equalsIgnoreCase("stochstr"))
			return STOCHSTR;

		if (str.equalsIgnoreCase("stochdat"))
			return STOCHDAT;

		if (str.equalsIgnoreCase("seed"))
			return SEED;

		return UNKNOWN;
	}

	
	public simsql.runtime.Attribute getPhysicalRealization() {

		switch (type) {
		case INTEGER:
			return new simsql.runtime.IntAttribute();
		case DOUBLE:
			return new simsql.runtime.DoubleAttribute();
		case STRING:
			return new simsql.runtime.StringAttribute();
		case DATE:
			return new simsql.runtime.StringAttribute();
		case STOCHINT:
			return new simsql.runtime.IntArrayAttribute();
		case STOCHDBL:
			return new simsql.runtime.DoubleArrayAttribute();
		case STOCHSTR:
			return new simsql.runtime.StringArrayAttribute();
		case STOCHDAT:
			return new simsql.runtime.StringArrayAttribute();
		case SEED:
			return new simsql.runtime.SeedAttribute();
		}

		throw new RuntimeException(
				"tried to convert an attribute type I didn't recognize!");
	}
	

	public String getTypeStr(boolean isRandom) {
		if (!isRandom)
			return (getTypeStr());

		switch (type) {
		case INTEGER:
			return ("stochint");
		case DOUBLE:
			return ("stochdbl");
		case STRING:
			return ("stochstr");
		case DATE:
			return ("stochdat");
		case STOCHINT:
			return ("stochint");
		case STOCHDBL:
			return ("stochdbl");
		case STOCHSTR:
			return ("stochstr");
		case STOCHDAT:
			return ("stochdat");
		case SEED:
			return ("seed");
		case BOTTOM:
			return ("bottom");
		default:
			return ("unknown");
		}
	}

	/** Returns a copy of this object */
	public Attribute clone() {
		return new Attribute(name, type, relName, isRandom);
	}

	public int getTypeSize(int type) {
		int size = 64; // default
		switch (type) {
		case INTEGER:
			size = 32;
			break;

		case DOUBLE:
			size = 64;
			break;

		case STRING:
			size = 320;
			break;

		case DATE:
			size = 320;
			break;

		case STOCHINT:
			size = 32;
			break;

		case STOCHDBL:
			size = 64;
			break;

		case STOCHSTR:
			size = 320;
			break;

		case STOCHDAT:
			size = 320;
			break;

		case SEED:
			size = 64;
			break;

		case BOTTOM:
			size = 64;
			break;

		case MULTITYPE:
			size = 64;
			break;

		case TOTALTYPE:
			size = 64;
			break;

		case UNKNOWN:
			size = 64;
			break;

		case HYBRID:
			size = 64;
			break;

		}

		return size;
	}
	
	public static int computeTypeSize(int type) {
		int size = 64; // default
		switch (type) {
		case INTEGER:
			size = 32;
			break;

		case DOUBLE:
			size = 64;
			break;

		case STRING:
			size = 320;
			break;

		case DATE:
			size = 320;
			break;

		case STOCHINT:
			size = 32;
			break;

		case STOCHDBL:
			size = 64;
			break;

		case STOCHSTR:
			size = 320;
			break;

		case STOCHDAT:
			size = 320;
			break;

		case SEED:
			size = 64;
			break;

		case BOTTOM:
			size = 64;
			break;

		case MULTITYPE:
			size = 64;
			break;

		case TOTALTYPE:
			size = 64;
			break;

		case UNKNOWN:
			size = 64;
			break;

		case HYBRID:
			size = 64;
			break;

		}

		return size;
	}

}
