/**
 * 
 */
package simsql.compiler; // package mcdb.compiler.logicPlan.logicOperator.statisticsOperator;

import java.util.ArrayList;
import java.util.HashMap;



// import mcdb.catalog.Attribute;
// import mcdb.catalog.Catalog;
// import mcdb.catalog.VGFunction;
// import mcdb.compiler.logicPlan.logicOperator.mathOperator.*;
// import mcdb.compiler.parser.astVisitor.TypeChecker;
// import mcdb.compiler.parser.expression.util.FinalVariable;


/**
 * @author Bamboo
 *
 */
public class TypeInference {
	private HashMap<String, Integer> attributeTypeMap;
	private MathOperator tempOperator;
	private Catalog catalog;
	
	
	public TypeInference(HashMap<String, Integer> attributeTypeMap,
			MathOperator tempOperator) {
		super();
		this.attributeTypeMap = attributeTypeMap;
		this.tempOperator = tempOperator;
		catalog = SimsqlCompiler.catalog;
	}
	
	public int getType() throws Exception
	{
		return getType(tempOperator);
	}
	
	public int getType(MathOperator mathOperator) throws Exception
	{
		if(mathOperator instanceof AggregateOperator)
		{
			return getTypeInAggregateOperator((AggregateOperator)mathOperator);
		}
		else if(mathOperator instanceof ArithmeticOperator)
		{
			return getTypeInArithmeticOperator((ArithmeticOperator)mathOperator);
		}
		else if(mathOperator instanceof ColumnOperator)
		{
			return getTypeInColumnOperator((ColumnOperator)mathOperator);
		}
		else if(mathOperator instanceof DateOperator)
		{
			return getTypeInDateOperator((DateOperator)mathOperator);
		}
		else if(mathOperator instanceof EFunction)
		{
			return getTypeInEFunction((EFunction)mathOperator);
		}
		else if(mathOperator instanceof FunctionOperator)
		{
			return getTypeInFunctionOperator((FunctionOperator)mathOperator);
		}
		else if(mathOperator instanceof PredicateToMathWrapper)
		{
			return Attribute.MULTITYPE;
		}
		else if(mathOperator instanceof NumberOperator)
		{
			return getTypeInNumberOperator((NumberOperator)mathOperator);
		}
		else if(mathOperator instanceof SetOperator)
		{
			return getTypeInSetOperator((SetOperator)mathOperator);
		}
		else if(mathOperator instanceof StarOperator)
		{
			return getTypeInStarOperator((StarOperator)mathOperator);
		}
		else if(mathOperator instanceof StringOperator)
		{
			return getTypeInStringOperator((StringOperator)mathOperator);
		}
		else
		{
			return Attribute.INTEGER;
		}
	}
	
	/**
	 * @param mathOperator
	 * @return
	 */
	private int getTypeInStringOperator(StringOperator mathOperator) {
		return Attribute.STRING;
	}

	/**
	 * @param mathOperator--------------------------*********************------------------------------
	 * @return
	 */
	private int getTypeInStarOperator(StarOperator mathOperator) {
		return Attribute.DOUBLE;
	}

	/**
	 * @param mathOperator
	 * @return
	 */
	private int getTypeInSetOperator(SetOperator mathOperator)  throws Exception{
		ArrayList<MathOperator> elementList = mathOperator.getElmentList();
		int elementTypes[] = new int[elementList.size()];
		
		for(int i = 0; i < elementList.size(); i++)
		{
			elementTypes[i] = getType(elementList.get(i));
		}
		
		return getType(elementTypes);
	}

	/**
	 * @param mathOperator
	 * @return
	 */
	private int getTypeInNumberOperator(NumberOperator mathOperator) {
		return mathOperator.type;
	}

	/**
	 * @param mathOperator
	 * @return
	 */
	private int getTypeInFunctionOperator(FunctionOperator mathOperator) throws Exception {
		String functionName = mathOperator.getFunctionName();
		Function func = catalog.getFunction(functionName);
		ArrayList<MathOperator>  parameterList = mathOperator.getParameterList();
		Attribute attribute = func.getOutputAtt();
		
		int elementTypes[] = new int[parameterList.size()];
		
		for(int i = 0; i < parameterList.size(); i++)
		{
			elementTypes[i] = getType(parameterList.get(i));
		}
		
		int parameterType = getType(elementTypes);
		
		int outType;
		
		if(functionName.equals("case"))
		{
			outType = elementTypes[elementTypes.length-1];
		}
		else
		{
			outType = attribute.getType();
		}
		
		
		if(isRandomTYpe(parameterType))
		{
			if(outType == Attribute.SEED)
			{
				return Attribute.SEED;
			}
			else if(outType == Attribute.STOCHDAT)
			{
				return Attribute.STOCHDAT;
			}
			else if(outType == Attribute.STOCHDBL)
			{
				return Attribute.STOCHDBL;
			}
			else if(outType == Attribute.STOCHINT)
			{
				return Attribute.STOCHINT;
			}
			else if(outType == Attribute.DATE)
			{
				return Attribute.STOCHDAT;
			}
			else if(outType == Attribute.STRING)
			{
				return Attribute.STRING;
			}
			else if(outType == Attribute.DOUBLE)
			{
				return Attribute.STOCHDBL;
			}
			else if(outType == Attribute.INTEGER)
			{
				return Attribute.STOCHINT;
			}
			else
				return Attribute.HYBRID;
		}
		else
		{
			return outType;
		}
	}

	/**
	 * @param mathOperator--------------------------*********************------------------------------
	 * @return
	 */
	private int getTypeInEFunction(EFunction mathOperator) {
		return Attribute.HYBRID;
	}

	/**
	 * @param mathOperator
	 * @return
	 */
	private int getTypeInDateOperator(DateOperator mathOperator) {
		return Attribute.DATE;
	}

	/**
	 * @param mathOperator
	 * @return
	 */
	private int getTypeInColumnOperator(ColumnOperator mathOperator) {
		return attributeTypeMap.get(mathOperator.getColumnName());
	}

	/**
	 * @param mathOperator
	 * @return
	 */
	private int getTypeInArithmeticOperator(ArithmeticOperator mathOperator)  throws Exception{
		MathOperator left = mathOperator.getLeft();
		MathOperator right = mathOperator.getRight();
		
		int leftType = getType(left);
		int rightType = getType(right);
		
		int typeArray[] = new int[2];
		typeArray[0] = leftType;
		typeArray[1] = rightType;
		
		if(typeArray[0] == Attribute.INTEGER && typeArray[0] == Attribute.INTEGER)
		{
			return Attribute.DOUBLE;
		}
		else
		{
			return getType(typeArray);
		}
	}

	/**
	 * @param mathOperator
	 * @return
	 */
	private int getTypeInAggregateOperator(AggregateOperator mathOperator)  throws Exception{
		/*
		 * The "type" is defined in Class expression.util.FinalVariable
		 *  public static final int AVG = 0;
		 *	public static final int SUM = 1;
		 *	public static final int COUNT = 2;
		 *	public static final int COUNTALL = 3;
		 *	public static final int MIN = 4;
		 *	public static final int MAX = 5;
		 *	public static final int VARIANCE = 6;
		 */
		int type = mathOperator.getType();
		
		/*
		 *  The "operator" can be:
		 *  	Expression
		 *  	Identifier (Column)
		 */
		MathOperator childOperator = mathOperator.getChildOperator();
		int childType = getType(childOperator);
		switch(type)
		{
			case FinalVariable.AVG:
			case FinalVariable.SUM:
			case FinalVariable.MIN:
			case FinalVariable.MAX:
			case FinalVariable.VARIANCE:
			case FinalVariable.STDEV:
				return childType;
				
			case FinalVariable.COUNT:
			case FinalVariable.COUNTALL:
				if(isRandomTYpe(childType))
				{
					return Attribute.STOCHINT;
				}
				else
				{
					return Attribute.INTEGER;	
				}
		}
		
		return Attribute.DOUBLE;
	}

	public HashMap<String, Integer> getAttributeTypeMap() {
		return attributeTypeMap;
	}
	
	
	public void setAttributeTypeMap(HashMap<String, Integer> attributeTypeMap) {
		this.attributeTypeMap = attributeTypeMap;
	}
	
	
	public MathOperator getTempOperator() {
		return tempOperator;
	}
	
	
	public void setTempOperator(MathOperator tempOperator) {
		this.tempOperator = tempOperator;
	}
	
	public int getType(int []types)
	{
		int typeNum[] = new int[Attribute.TOTALTYPE];
		for(int i = 0; i < typeNum.length; i++)
		{
			typeNum[i] = 0;
		}
		
		for(int i = 0; i < types.length; i++)
		{
			typeNum[types[i]] ++;
		}
		
		if(typeNum[Attribute.SEED] != 0)
		{
			return Attribute.SEED;
		}
		else if(typeNum[Attribute.STOCHDAT] != 0)
		{
			return Attribute.STOCHDAT;
		}
		else if(typeNum[Attribute.STOCHDBL] != 0)
		{
			return Attribute.STOCHDBL;
		}
		else if(typeNum[Attribute.STOCHINT] != 0)
		{
			if(typeNum[Attribute.DOUBLE] != 0)
			{
				return Attribute.STOCHDBL;
			}
			else
			{
				return Attribute.STOCHINT;
			}
		}
		else if(typeNum[Attribute.DATE] != 0)
		{
			return Attribute.DATE;
		}
		else if(typeNum[Attribute.STRING] != 0)
		{
			return Attribute.STRING;
		}
		else if(typeNum[Attribute.DOUBLE] != 0)
		{
			return Attribute.DOUBLE;
		}
		else if(typeNum[Attribute.INTEGER] != 0)
		{
			return Attribute.INTEGER;
		}
		else
			return Attribute.HYBRID;
	}
	
	public boolean isRandomTYpe(int type)
	{
		if(type == Attribute.STOCHDAT ||
				type == Attribute.STOCHDBL ||
				type == Attribute.STOCHINT)
		{
			return true;
		}
		else
		{
			return false;
		}
	}
	
}
