#include "VGFunction.h"
#include "lib/InverseGaussian.h"
#include "lib/InverseGaussian.cc"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <stdio.h>
#include <vector>
#include <map>
#include <utility>
#include <math.h>
#include <iostream>
#include <gsl/gsl_errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  double *mean;
  double *shape;
};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  double *outValue;
};

int runningError;
void handler (const char * reason,
	      const char * file,
	      int line,
	      int gsl_errno)
{
  runningError = 1;
}




class InvGaussian: public VGFunction {
private:
  gsl_rng *rng;

  InverseGaussian *ig;

  double result;
  bool active;

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  InvGaussian() {

    // create the RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);
    runningError = -1;

    ig = new InverseGaussian();
  }

  ~InvGaussian()
  {
    // destroy the RNG
    gsl_rng_free(rng);

    delete ig;
  }


  void initializeSeed(long seed) {

    // set the seed
    gsl_rng_set(rng, seed);
  }

  /** Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). */
  void finalizeTrial() {
    active = true;
  }


  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {

    ig->ClearParameters();
    active = false;
  }

  //take the parameters for vgfunction
  void takeParams(RecordIn &input) {

    // prior
    if (input.mean != NULL) {
      ig->SetMean(*input.mean);
    }

    if (input.shape != NULL) {
      ig->SetShape(*input.shape);
    }

    active = true;
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) {

    if(!active || ig->GetMean() <= 0 || ig->GetShape() <= 0) {
      return 0;
    }

		
    result = ig->Sample(rng);
    output.outValue = &result;
		
    active = false;
    return 1;
  }

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){2, {DOUBLE, DOUBLE}, {"mean", "shape"}, {1, 1}};
  }

  VGSchema outputSchema() {

    return (VGSchema){1, {DOUBLE}, {"outValue"}, {1}};
  }

  const char *getName() {
    return "InvGaussian";
  }
};


/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new InvGaussian());
}

void destroy(VGFunction *vgFunction) {
  delete (InvGaussian *)vgFunction;
}
