
package simsql.runtime;

/**
 * This class implements an Attribute that is stored as an array, 
 * where the value in one more more of the worlds is NULL
 */
import java.util.*;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.PrintStream;

public class ArrayAttributeWithNulls implements Attribute {
  
  // there is a one in every position that has a null
  private Bitstring theseOnesAreNull;
  
  // this is the array of values
  private Attribute myAttribute;
  
  // these are the machines to read/write the above two members
  static BitstringIOMachine bitstringMachine = new BitstringIOMachine ();
  static AttributeIOMachine attributeMachine = new AttributeIOMachine ();

  public void recycle() {
  }

  public ArrayAttributeWithNulls () {}
  
  public Attribute readSelfFromTextStream (BufferedReader readFromMe) throws IOException {
    throw new IOException ("It makes no sense to try to read text into an ArrayAttributeWithNulls object");
  }

  public void writeSelfToTextStream (BufferedWriter writeToMe) throws IOException {
    myAttribute.writeSelfToTextStream (theseOnesAreNull, writeToMe);
  }

  public void writeSelfToTextStream (Bitstring theseAreNull, BufferedWriter writeToMe) throws IOException {

    // in this case, we write a null if indicated by either theseAreNull or theseOnesAreNull
    myAttribute.writeSelfToTextStream (theseOnesAreNull.or (theseAreNull), writeToMe);
  }


  public int writeSelfToStream (DataOutputStream writeToMe) throws IOException {
    int returnVal = bitstringMachine.writeToStream (theseOnesAreNull, writeToMe);
    return returnVal + attributeMachine.writeToStream (myAttribute, writeToMe);
  }
  
  public Attribute removeNulls () {
    return myAttribute;
  }
  
  public long getHashCode () {
    throw new RuntimeException ("Can't hash an array attribute!");  
  }

  public int getSize () {
   return myAttribute.getSize ();
  }
  
  public byte[] getValue (int whichMCIter) {
    if (!theseOnesAreNull.getValue (whichMCIter)) {
      return myAttribute.getValue (whichMCIter); 
    } else {
      throw new RuntimeException ("you tried to get a value that was NULL");
    }
  }
  
  public byte[] getValue (int whichMCIter, AttributeType castTo) {
    if (!theseOnesAreNull.getValue (whichMCIter)) {
	return myAttribute.getValue (whichMCIter, castTo); 
    } else {
      throw new RuntimeException ("you tried to get a value that was NULL");
    }
  }
  
  public AttributeType getType (int whichMCIter) {
   if (!theseOnesAreNull.getValue (whichMCIter)) {
      return myAttribute.getType (whichMCIter); 
    } else {
      return AttributeType.NULL;
    }
  }
  
  public int readSelfFromStream (DataInputStream readFromMe) throws IOException {
    WriteableInt numBytes = new WriteableInt ();
    theseOnesAreNull = bitstringMachine.readFromStream (readFromMe, numBytes);
    int totRead = (int) numBytes.get ();
    myAttribute = attributeMachine.readFromStream (readFromMe, numBytes, numBytes); 
    return totRead + (int) numBytes.get ();
  }
  
  public ArrayAttributeWithNulls (Bitstring nullOnes, Attribute fromMe) {
    theseOnesAreNull = nullOnes.or (fromMe.isNull ());
    myAttribute = fromMe;
  }
  
  
  public Bitstring isNull () {
    return theseOnesAreNull;
  }
  
  public Attribute setNull (Bitstring theseOnes) {
    Bitstring temp = theseOnes.or (theseOnesAreNull);
    if (temp.allAreTrue ()) {
      return NullAttribute.NULL;
    } else {
      return new ArrayAttributeWithNulls (temp, this);
    }
  }

  public HashMap<Attribute, Bitstring> split () {
   HashMap<Attribute, Bitstring> splits = myAttribute.split();
 for(Attribute att : splits.keySet()) {   
   splits.put(att, splits.get(att).and(theseOnesAreNull.not()));              
 }  
   return splits;
  }
  
  public Attribute add (Attribute me) {
    Attribute res = myAttribute.add (me);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute add (long addThisIn) {
    Attribute res = myAttribute.add (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute addR (long addThisIn) {
    Attribute res = myAttribute.addR (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute add (long [] addThisIn) {
    Attribute res = myAttribute.add (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute addR (long [] addThisIn) {
    Attribute res = myAttribute.addR (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute add (String [] addThisIn) {
    Attribute res = myAttribute.add (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute addR (String [] addThisIn) {
    Attribute res = myAttribute.addR (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute add (String addThisIn) {
    Attribute res = myAttribute.add (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute addR (String addThisIn) {
    Attribute res = myAttribute.addR (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute add (double [] addThisIn) {
    Attribute res = myAttribute.add (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute addR (double [] addThisIn) {
    Attribute res = myAttribute.addR (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute add (double addThisIn) {
    Attribute res = myAttribute.add (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute addR (double addThisIn) {
    Attribute res = myAttribute.addR (addThisIn);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute subtract (Attribute subtractThisOut) {
    Attribute res = myAttribute.subtract (subtractThisOut);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute subtract (long subtractThisOut) {
    Attribute res = myAttribute.subtract (subtractThisOut);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute subtractR (long subtractFromMe) {
    Attribute res = myAttribute.subtractR (subtractFromMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute subtract (long [] subtractMeOut) {
    Attribute res = myAttribute.subtract (subtractMeOut);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute subtractR (long [] subtractFromMe) {
    Attribute res = myAttribute.subtractR (subtractFromMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute subtract (double [] subtractThisOut) {
    Attribute res = myAttribute.subtract (subtractThisOut);
    return res.setNull (theseOnesAreNull); 
  }
  
  public Attribute subtractR (double [] subtractFromMe) {
    Attribute res = myAttribute.subtractR (subtractFromMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute subtract (double subtractThisOut) {
    Attribute res = myAttribute.subtract (subtractThisOut);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute subtractR (double subtractFromThis) {
    Attribute res = myAttribute.subtractR (subtractFromThis);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute multiply (Attribute byMe) {
    Attribute res = myAttribute.multiply (byMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute multiply (long byMe) {
    Attribute res = myAttribute.multiply (byMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute multiply (double byMe) {
    Attribute res = myAttribute.multiply (byMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute multiply (long [] byMe) {
    Attribute res = myAttribute.multiply (byMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute multiply (double [] byMe) {
    Attribute res = myAttribute.multiply (byMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute divide (Attribute byMe) {
    Attribute res = myAttribute.divide (byMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute divide (long byMe) {
    Attribute res = myAttribute.divide (byMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute divide (double byMe) {
    Attribute res = myAttribute.divide (byMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute divide (long [] byMe) {
    Attribute res = myAttribute.divide (byMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute divide (double [] byMe) {
    Attribute res = myAttribute.divide (byMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute divideR (long divideMe) {
    Attribute res = myAttribute.divideR (divideMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute divideR (double divideMe) {
    Attribute res = myAttribute.divideR (divideMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute divideR (long [] divideMe) {
    Attribute res = myAttribute.divideR (divideMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Attribute divideR (double [] divideMe) {
    Attribute res = myAttribute.divideR (divideMe);
    return res.setNull (theseOnesAreNull);
  }
  
  public Bitstring equals (Attribute me) {
    Bitstring res = myAttribute.equals (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring equals (long me) {
    Bitstring res = myAttribute.equals (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring equals (double me)  {
    Bitstring res = myAttribute.equals (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring equals (String me)  {
    Bitstring res = myAttribute.equals (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring equals (long [] me)  {
    Bitstring res = myAttribute.equals (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring equals (double [] me)  {
    Bitstring res = myAttribute.equals (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring equals (String [] me)  {
    Bitstring res = myAttribute.equals (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring notEqual (Attribute me) {
    Bitstring res = myAttribute.notEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring notEqual (long me) {
    Bitstring res = myAttribute.notEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring notEqual (double me)  {
    Bitstring res = myAttribute.notEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring notEqual (String me)  {
    Bitstring res = myAttribute.notEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring notEqual (long [] me)  {
    Bitstring res = myAttribute.notEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring notEqual (double [] me)  {
    Bitstring res = myAttribute.notEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring notEqual (String [] me)  {
    Bitstring res = myAttribute.notEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThan (Attribute me) {
    Bitstring res = myAttribute.greaterThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThan (long me) {
    Bitstring res = myAttribute.greaterThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThan (double me) {
    Bitstring res = myAttribute.greaterThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThan (String me) {
    Bitstring res = myAttribute.greaterThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThan (long [] me) {
    Bitstring res = myAttribute.greaterThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThan (double [] me) {
    Bitstring res = myAttribute.greaterThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThan (String [] me)  {
    Bitstring res = myAttribute.greaterThan (me);
    return res.setUnknown (theseOnesAreNull);
  }

  public Bitstring lessThan (Attribute me) {
    Bitstring res = myAttribute.lessThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThan (long me) {
    Bitstring res = myAttribute.lessThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThan (double me) {
    Bitstring res = myAttribute.lessThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThan (String me) {
    Bitstring res = myAttribute.lessThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThan (long [] me) {
    Bitstring res = myAttribute.lessThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThan (double [] me)  {
    Bitstring res = myAttribute.lessThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThan (String [] me) {
    Bitstring res = myAttribute.lessThan (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThanOrEqual (Attribute me) {
    Bitstring res = myAttribute.greaterThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThanOrEqual (long me) {
    Bitstring res = myAttribute.greaterThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThanOrEqual (double me) {
    Bitstring res = myAttribute.greaterThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThanOrEqual (String me) {
    Bitstring res = myAttribute.greaterThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThanOrEqual (long [] me) {
    Bitstring res = myAttribute.greaterThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThanOrEqual (double [] me) {
    Bitstring res = myAttribute.greaterThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring greaterThanOrEqual (String [] me)  {
    Bitstring res = myAttribute.greaterThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }

  public Bitstring lessThanOrEqual (Attribute me) {
    Bitstring res = myAttribute.lessThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThanOrEqual (long me) {
    Bitstring res = myAttribute.lessThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThanOrEqual (double me) {
    Bitstring res = myAttribute.lessThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThanOrEqual (String me) {
    Bitstring res = myAttribute.lessThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThanOrEqual (long [] me) {
    Bitstring res = myAttribute.lessThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThanOrEqual (double [] me)  {
    Bitstring res = myAttribute.lessThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  
  public Bitstring lessThanOrEqual (String [] me) {
    Bitstring res = myAttribute.lessThanOrEqual (me);
    return res.setUnknown (theseOnesAreNull);
  }
  

  public void injectSelf(Function f) {

      // inject the attribute contents
      myAttribute.injectSelf(f);

      // and the bitstring.
      f.injectBitstring(theseOnesAreNull);
  }

  public boolean allAreEqual() {
      return myAttribute.allAreEqual();
  }

  public String print(int maxLen) {
      return print(maxLen, theseOnesAreNull);
  }

  public String print(int maxLen, Bitstring theseAreNull) {
      return myAttribute.print(maxLen, theseAreNull.or(theseOnesAreNull));
  }

    public Attribute getSingleton() {
	return myAttribute.getSingleton();
    }
}
