
package simsql.runtime;

import org.apache.hadoop.io.serializer.Deserializer;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.FileNotFoundException;

/**
 * This is used to deserialize record keys so thay can be processed
 * by a reducer
 */
class RecordKeyDeserializer implements Deserializer <WritableKey> {
  
  // this is the stream we deserialize from
  private InputStream getFromMe;
  
  // this is where we store the data
  byte [] dataHere = new byte [10];
  
  public void open (InputStream in) throws IOException {
    getFromMe = in;
  }
  
  public void open (String pathToWriteTo) {
    try {
      FileInputStream inFileStream = new FileInputStream (pathToWriteTo);
      getFromMe = new BufferedInputStream (inFileStream, 32768);
    } catch (FileNotFoundException e) {
      throw new RuntimeException ("for some reason I could not create the file you asked for"); 
    }
  }
  
  public void close () throws IOException {
    if (getFromMe != null)
      getFromMe.close ();
  }
  
  public WritableKey deserialize (WritableKey input) throws IOException {

      /***
      if (input instanceof Nothing)
	  return input;
      ***/

    // do the read
    int returnVal = getFromMe.read (dataHere);
    
    // if we didn't get 10 bytes, we have a problem
    if (returnVal != 10) {
      throw new IOException (); 
    }
    
    // get the key
    long newKey = 0;
    for (int i = 0; i < 8; i++) {
      newKey += ((long) (dataHere[i] & 0xFF)) << ((7 - i) * 8);
    }
    
    // get the code
    short code = 0;
    for (int i = 0; i < 2; i++) {
      code += ((short) (dataHere[i + 8] & 0xFF)) << ((1 - i) * 8);
    }
    
    // and create the object
    if (input == null) {
      return new RecordKey (newKey, code); 
    } else {
      input.set (newKey, code);
      return input;
    }
  }
  
}
