
package simsql.runtime;

import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import org.apache.hadoop.io.serializer.Serializer;
import java.io.IOException;
import java.io.FileNotFoundException;

/**
 * This is used to serialize RecordKey objects so they can be sorted
 * after they are created by the mapper
 */
class RecordKeySerializer implements Serializer <WritableKey> {
  
  private OutputStream writeToMe;
  
  public void close () throws IOException {
    if (writeToMe != null)
      writeToMe.close ();
  }
  
  public void open (OutputStream out) throws IOException {
    writeToMe = out; 
  }
  
  public void open (String pathToWriteTo) {
    try {
      FileOutputStream outFileStream = new FileOutputStream (pathToWriteTo);
      writeToMe = new BufferedOutputStream (outFileStream, 32768);
    } catch (FileNotFoundException e) {
      throw new RuntimeException ("for some reason I could not create the file you asked for"); 
    }
  }
    
  public void serialize (WritableKey serializeMe) throws IOException {

      /**
	 if (serializeMe instanceof Nothing)
	 return;
      **/

    // this is where we write the output
    byte [] outputArray = new byte [10];
    
    // serialize the key
    long key = serializeMe.getKey ();
    for (int i = 0; i < 8; i++) {
       outputArray[i] = (byte) ((key >>> ((7 - i) * 8)) & 0xFF);
    }
    
    // serialize the code
    short code = serializeMe.getTypeCode ();
    outputArray[8] = (byte) ((code >>> 8) & 0xFF);
    outputArray[9] = (byte) (code & 0xFF);
    
    // and do the write
    writeToMe.write (outputArray);
  }
}
