package simsql.runtime;


/* Encapsulates all the aggregation and pre-aggregation of
 * records. Essentially, the "innards" of the Aggregate Op.
 *
 * @author Luis.
 */
class Aggregator {
  
  // estimator for the size of our hash table.
  private CollectionSizeEstimator cse;
  private RecordHashTable table;

  // adapter
  private OutputAdapter writer;

  // latest key.
  private long lastKey = -1;
  
  /** Simple constructor. */
  public Aggregator(long memory, OutputAdapter writer) {

    table = new RecordHashTable(0.25, 1024);
    cse = new CollectionSizeEstimator(table.getCollectionWrapper(), memory);
    this.writer = writer;
  }

  /** Takes an input record for aggregation. */
  public void take(Record inRec) {

    // cast to aggregation record
    AggregatorRecord aggRec = null;
    try {
      aggRec = (AggregatorRecord) inRec;
    } catch (Exception e) {
	throw new RuntimeException("Could not cast selected record to the relevant type!", e);
    }

    Record[] preAggRecs = table.find (aggRec.getSecondaryHashKey());
    boolean foundIt = false;
    if (preAggRecs != null) {
      
      // if so, find the one that matches
      for (int i=0;i<preAggRecs.length;i++) {
	AggregatorRecord preAggRec = (AggregatorRecord)preAggRecs[i];
	
	if (preAggRec.isFromSameGroup(aggRec)) {

	  // if we found it, then consume the current record.
	  preAggRec.consume(aggRec);
	  foundIt = true;
	  break;
	}
      }
    }

    // did we not find it? then put it in the table.
    if (!foundIt) {      

      // but, first, check if a flush is due.
      if (cse.capacityExceeded() && lastKey != aggRec.getHashKey()) {
	
	// if so, flush -- this is OK
	flush();
      }

      table.add(aggRec, aggRec.getSecondaryHashKey());
      cse.updateEstimate();      

      // update the key.
      lastKey = aggRec.getHashKey();
    } else {

	// otherwise, recycle the consumed aggregator.
	lastKey = aggRec.getHashKey();
	aggRec.recycle();
    }

  }

  /** Flushes all the aggregated records we have. */
  public void flush() {

    // go through all the records in the table.
    for (Record xrec: table) {

      // give them to the adapter; it knows what to do.
      writer.writeRecord(xrec);
    }

    // clear the table
    table.clear();
  }
}