#ifndef _VGFUNCTIONLIB_H
#define _VGFUNCTIONLIB_H

// This must mirror the AttributeType enum from the Java classes.
enum VGType {
  NULLS,
  SEED,
  INT,
  DOUBLE,
  STRING
};

// Encapsulates the schema
struct VGSchema {
  int numAtts;
  VGType attTypes[16384];
  const char *attNames[16384];
  int isRandom[16384];
};

class RecordIn;
class RecordOut;

// The general abstract definition of a VG Function.
class VGFunction {
 public:

  // public functions
  virtual void initializeSeed(long seed) = 0;
  virtual void takeParams(RecordIn &in_rec) = 0;
  virtual int outputVals(RecordOut &out_rec) = 0;
  virtual void finalizeTrial() = 0;
  virtual void clearParams() = 0;
  virtual VGSchema inputSchema() = 0;
  virtual VGSchema outputSchema() = 0;
  virtual const char *getName() = 0;
};

// These methods are generated for each VGFunction type. 
extern "C" {
  VGFunction *create();
  void destroy(VGFunction *vg);
};

#endif // _VGFUNCTIONLIB_H
