
package simsql.runtime;

import org.apache.hadoop.mapreduce.Reducer;
import java.util.*;
import java.io.IOException;
import org.apache.hadoop.mapreduce.TaskInputOutputContext;

/**
 * The reducer for the selection/projection/scalarFunc operator.
 * To be invoked only when removing duplicates.
 */
class SelectionReducer extends ReducerWithPostProcessing {

  // adapter class for writing -- this one can be used by the mapper.
  public static class Writer implements OutputAdapter {

    private TaskInputOutputContext <?, ?, WritableKey, WritableValue> context;
    private Nothing dummy = new Nothing();

    public Writer(TaskInputOutputContext <?, ?, WritableKey, WritableValue> context) {
      this.context = context;
    }

    public void writeRecord(Record writeMe) {
      try {
	if (writeMe != null) {
	  context.write(dummy, writeMe);
	}
      } catch (Exception e) {
	throw new RuntimeException("Failed to write out record!", e);
      }
    }
  }

  // set with duplicates
  private Deduper dedup;

  // set up the reducer...
  public void setup(Context context) throws IOException, InterruptedException {

    // determine if we are removing duplicates.
    boolean removeDuplicates = context.getConfiguration().getBoolean("simsql.removeDuplicates", false);
    dedup = new Deduper(1024L * 1024L * 10, new Writer(context), removeDuplicates);
  }
    
  public void reduce (RecordKey key, Iterable <RecordWrapper> values, Context context) {

    // go through all the input records -- that's it.
    for (RecordWrapper rw: values) {

      Record rec = (Record)rw.getWrappedRecord();
      dedup.take(rec);
    }
  }
}
