package simsql.runtime;

import java.util.*;

/**
 * Stub of the appendable selection operator. Template classes derive from this one
 * and implement the passInput(), and accepts() methods.
 *
 * @author Luis.
 */
abstract class PipelinedSelectionOp extends PipelinedOperator {

  // our writer adapter class for the duplicate remover.
  private class Writer implements OutputAdapter {
    public Writer() {
    }

    public void writeRecord(Record writeMe) {
      HashableRecord r = (HashableRecord)writeMe;
      if (r.hasSortAttribute()) {
	  r.setSortAttribute(r.getSortAttribute());
      }
      emit(r);
    }
  }

  private Deduper dedup;
  private Writer myWriter;

  public PipelinedSelectionOp(boolean removeDuplicates) {
    myWriter = new Writer();
    dedup = new Deduper(16L * (1024L * 1024L), myWriter, removeDuplicates);
  }

  // this is for the generated classes to implement.  it will just
  // cast the input record and run the selection and projection.
  protected abstract HashableRecord passInput(Record rec);

  public void process (Record me) {

    // run selection and production.
    HashableRecord hashedRec = passInput(me);
    if (hashedRec == null) {
      return;
    }

    // give it to the deduper.
    dedup.take(hashedRec);
  }
}