package simsql.runtime;
import java.io.*;


/**
 * Due to stupidity in hadoop, we can't just send generic "Record" objects on the output
 * stream from the Mapper.  The problem is that Record is an interface.  If we do send
 * object that implement this interface, it will error out with seomthing like:
 * 
 * Type mismatch in value from map: expected Record, recieved OutOrders
 * 
 * Even though OutOrders implements Record!  To get around this, the mapper needs to wrap
 * the record inside of a RecordWrapper object, and then send THAT along
 * 
 */ 
public class RecordWrapper implements WritableValue {
  
  private Record holdMe;
  
  public Record getWrappedRecord () {
    return holdMe;
  }
  
  public RecordWrapper (Record useMe) {
    holdMe = useMe;
  }
  
  public RecordWrapper () {
    holdMe = null; 
  }

  // a 'set' method so that we can recycle instances of this class.
  public void set(Record useMe) {
    holdMe = useMe;
  }


  public int writeSelfToStream (DataOutputStream writeToMe) throws IOException {
    return holdMe.writeSelfToStream(writeToMe);
  }

  public int readSelfFromStream (DataInputStream readFromMe) throws IOException {
    return holdMe.readSelfFromStream(readFromMe);
  }

}
