package simsql.shell;
import java.io.*;
import java.lang.*;

// this interface defines a container for the statistics of a given
// table
public interface TableStatistics <RecordIn> {

  // clears the structure, back to its default 'empty' values.
  void clear();

  // takes an input record and updates the statistics.
  void take(RecordIn inRec);

  // consumes another set of statistics from the same table, associated
  // with another partition.
  void consume(TableStatistics<RecordIn> inTable);

  // loads the statistics from a file.
  void load(String path) throws IOException, ClassNotFoundException;

  // saves the contents to a file.
  void save(String path) throws IOException;

  // returns the total number of tuples in a given table.
  long numTuples();

  // returns the number of attributes.
  int numAttributes();

  // returns the number of unique values for any given attribute.
  long numUniques(int att);
}